-- Knowledge base core schema

-- Copyright (C) 2018, 2019 Nathan Matthew Miller

-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <https://www.gnu.org/licenses/>.

-- Tables which store information about knowledge fragments and their
-- relationships.

-- Various settings. The KB's UUID value is one item stored in this table.
--
CREATE TABLE settings (
       name    TEXT NOT NULL,
       value   TEXT NOT NULL,
       section TEXT
);


-- When a foreign KB is imported, this table contains identifying information.
--
CREATE TABLE imported_kb (
       origin_id   INTEGER PRIMARY KEY,
       origin_name TEXT,
       uuid        TEXT
);


-- A KB entity is either a knowledge fragment (type = 'K') or a category
-- (type = 'C'). The "origin" column is used to differentiate between the owner's
-- content and content imported from ther KBs.
--
CREATE TABLE kb_entity (
       id       TEXT PRIMARY KEY,
       name     TEXT,
       created  TIMESTAMP DEFAULT ( datetime('now') ),
       modified TIMESTAMP DEFAULT ( datetime('now') ),
       type     TEXT DEFAULT 'K',
       origin   INTEGER DEFAULT 0
);


-- Entities can be related by a directed or undirected relationship.
--
CREATE TABLE relationship (
       parent   TEXT,
       child    TEXT,
       directed TEXT DEFAULT 'N',
       CONSTRAINT relationship_pk PRIMARY KEY (
           parent, child 
       ),
       CONSTRAINT r_parent_fk FOREIGN KEY
           ( parent ) REFERENCES kb_entity ( id ),
       CONSTRAINT r_child_fk FOREIGN KEY
           ( child ) REFERENCES kb_entity ( id )
);

CREATE TRIGGER no_self_references
BEFORE INSERT ON relationship
WHEN EXISTS ( SELECT 1 WHERE NEW.parent = NEW.child )
BEGIN
    SELECT RAISE( ROLLBACK, 'Circular relationships not allowed' );
END;


-- A category can contain other entities, whether they are knowledge fragments
-- or other categories. An ordering can also be chosen for the entities.
--
CREATE TABLE category (
       id       TEXT,
       contains TEXT,
       ordering INTEGER DEFAULT 0, -- the default means "no ordering"
       CONSTRAINT category_fk1 FOREIGN KEY ( id )
           REFERENCES kb_entity ( id ),
       CONSTRAINT category_fk2 FOREIGN KEY ( contains )
           REFERENCES kb_entity ( id )
);

CREATE INDEX category_idx1
ON category ( contains );

CREATE INDEX category_idx2
ON category ( id );

-- Prevent any duplication of entities in the category.
--
CREATE TRIGGER no_duplicate_entries
BEFORE INSERT ON category
WHEN EXISTS ( SELECT 1 FROM category c
    WHERE c.id = NEW.id AND c.contains = NEW.contains )
BEGIN
    SELECT RAISE( ROLLBACK, 'Entity already present in the category' );
END;


-- Storage for knowledge fragments. These are the basic items in the knowledge
-- base.
--
CREATE TABLE k_frag (
       id           TEXT PRIMARY KEY,
       content      TEXT,
       binary       BLOB,
       binary_type  TEXT,
       content_type TEXT
);


--
-- Helpful views
--

CREATE VIEW frag AS
SELECT f.id, e.name, e.created, e.modified, e.origin,
       f.content, f.binary, f.content_type, f.binary_type
   FROM k_frag AS f JOIN kb_entity AS e
        ON f.id = e.id;


