# Core classes and routines for the knowledge base.
#
# This part of the application deals with all interactions with the database.

# Copyright (C) 2018, 2019 Nathan Matthew Miller

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import uuid
import os
import os.path
import configparser
import sqlite3
import platform


# A knowledge fragment
#
class KFrag:

    def __init__( self ):
        return


    # Properties
    #  id
    #  name
    #  created
    #  modified
    #  type
    #  origin
    #
    #  content, content_type
    #  binary, binary_type

    
# A category
#
class Category:

    def __init__( self ):
        return

    # Properties
    #  id
    #  name
    #  created
    #  modified
    #  type
    #  origin
    #
    #  contains - an array of tuples (id, ordering)


class Core:

    def __init__( self, db_file=None ):

        self.REG_PATH = r'Software\Exocortex'
        
        db_file_to_open = db_file
        
        if db_file == None:

            if platform.system() == "Windows":

                import winreg
                registry_key = winreg.OpenKey(
                    winreg.HKEY_CURRENT_USER, self.REG_PATH, 0,
                    winreg.KEY_READ )
                (value, regtype) = winreg.QueryValueEx( registry_key,
                                                        r'main' )
                winreg.CloseKey( registry_key )

                db_file_to_open = value
            else:
                        
                config_file = os.path.expanduser( "~/.exocortex" )

                if not os.access( config_file, os.R_OK ):
                    raise RuntimeError(
                        f"Configuration file {config_file} not found or not readable\n" )

                config = configparser.ConfigParser()
                config.read( config_file )

                db_file_to_open = os.path.expanduser( config['databases']['main'] )        

        print( f"Opening DB {db_file_to_open} ...")
        self.conn = sqlite3.connect( db_file_to_open )
        self.conn.execute( 'PRAGMA foreign_keys = ON' )
        self.conn.row_factory = sqlite3.Row
        return
    

    def close( self ):

        self.conn.close()
        return


    # UUID4 IDs are random and clashes rare, but let's not bet on it.
    #
    def create_entity_uuid( self ):

        not_unique = True
        limit = 5
        
        while not_unique:

            uuid = uuid_create();
            with self.conn as c:
                cur = c.execute(
                    '''SELECT 1 FROM kb_entity WHERE id = ?''',
                    (uuid,) )
                if cur.fetchone() == None:
                    not_unique = False
            limit -= 1
            if limit == 0:
                raise RuntimeError( "create_entity_uuid limit test reached" )
        return uuid

    
    # Create a knowledge fragment
    # Requirements: at least one of the content or binary arguments must be
    #  supplied.
    # Return: the UUID of the new fragment.
    #
    def kf_create(
               self, name = None,
               content = None, content_type = 'text/plain',
               binary  = None, binary_type  = 'application/octet-stream',
               created = None, modified = None ):

        uuid = self.create_entity_uuid()
        complete = False

        cols   = [ 'id' ]
        values = [ uuid ]
        
        if content != None:
            cols.append( 'content' );      values.append( content )
            cols.append( 'content_type' ); values.append( content_type )
            complete = True

        if binary != None:
            cols.append( 'binary' );      values.append( binary )
            cols.append( 'binary_type' ); values.append( binary_type )
            complete = True

        if not complete:
            raise RuntimeError( 'kf_create: neither content or binary provided' )

        e_cols = [ 'id', 'name', 'type' ]
        e_vals = [ uuid, name, 'K' ]
        
        if created != None:
            e_cols.append( 'created' ); e_vals.append( created )

        if modified != None:
            e_cols.append( 'modified' ); e_vals.append( modified )

        sq2 = 'INSERT INTO kb_entity (' + ','.join( e_cols ) + \
              ') VALUES (' + ','.join( ['?'] * len(e_vals) ) + ')'
        
        sql = 'INSERT INTO k_frag (' + ','.join( cols ) + \
              ') VALUES (' + ','.join( ['?'] * len(values) ) + ')'
              
        with self.conn as c:
            c.execute( sq2, tuple(e_vals) )
            c.execute( sql, tuple(values) )
        
        return uuid


    # Update a knowledge fragment. Supply the kf/entity ID and one or both
    # of the content and binary values.
    #
    def kf_update( self, entity_id, name = None, content = None, binary = None ):

        complete = False

        cols   = []
        values = []
        
        if content:
            cols.append( 'content' ); values.append( content )
            complete = True

        if binary:
            cols.append( 'binary' );  values.append( binary )
            complete = True

        if not complete:
            raise RuntimeError( 'kf_update: neither content nor binary provided' )

        values.append( entity_id )
        
        update_cols = []
        for col in cols:
            update_cols.append( col + "=?" )
            
        sql = 'UPDATE k_frag SET ' + ','.join(update_cols) + \
              ' WHERE id = ?'

        sql2 = '''
               UPDATE kb_entity SET name=?, modified=datetime('now')
                  WHERE id = ?
               '''
        with self.conn as c:
            cur1 = c.execute( sql, tuple(values) )
            cur2 = c.execute( sql2, (name, entity_id) )

            rowcount = cur1.rowcount + cur2.rowcount
            if rowcount != 2:
                c.rollback()
                raise RuntimeError(
                    f'kf_update: rowcount was {rowcount} instead of 2' )
        return
    

    # Create a new category and return its identifying UUID.
    # Categories must be named, while knowledge fragments do not.
    #
    def category_create( self, name ):

        uuid = self.create_entity_uuid()
        
        with self.conn as c:

            c.execute( '''INSERT INTO kb_entity (id, name, type)
                VALUES (?,?,'C')''', (uuid, name) )
        return uuid


    # Add an entity to a category.
    #
    def category_add( self, category_id, entity_id, ordering=0 ):

        with self.conn as c:

            c.execute( '''INSERT INTO category (id, contains, ordering )
                VALUES (?,?,?)''', (category_id, entity_id, ordering) )

            c.execute( '''
               UPDATE kb_entity SET modified=datetime('now')
                  WHERE id = ?
              ''', (category_id,) )
        return


    # Remove an entity from a category.
    #
    def category_remove( self, category_id, entity_id ):

        with self.conn as c:

            cur = c.execute( '''DELETE FROM category WHERE id = ? AND
                contains = ?''', (category_id, entity_id) )
            rowcount = cur.rowcount
            if rowcount != 1:
                cur.rollback()
                raise RuntimeError(
                    f'category_remove: rowcount was {rowcount} instead of 1' )
            
            c.execute( '''
               UPDATE kb_entity SET modified=datetime('now')
                  WHERE id = ?
              ''', (category_id,) )
        return


    # Delete a category. This action does not delete any entities which
    # the category contains.
    #
    def category_delete( self, category_id ):

        with self.conn as c:

            cur = c.execute( '''DELETE FROM category WHERE id = ?''',
                             (category_id,) )

            rowcount = cur.rowcount
            if rowcount != 0:
                cur.rollback()
                raise RuntimeError(
                    f'category_delete: rowcount was {rowcount} instead of >0' )
        return


    # A category's contents are changed using other methods. This method
    # only updates the name and modification time.
    #
    def category_update( self, entity_id, name ):

        sql = '''
               UPDATE kb_entity SET name=?, modified=datetime('now')
                  WHERE id = ?
              '''
        with self.conn as c:
            cur = c.execute( sql, (name, entity_id) )

            rowcount = cur.rowcount
            if rowcount != 1:
                cur.rollback()
                raise RuntimeError(
                    f'category_update: rowcount was {rowcount} instead of 1' )
        return


    ##
    ## TODO: Ordering items in a category. Do it singly or use a list of
    ##       UUID's to set the order?
    ##

    # Create a relationship between two entities. If the directed argument
    # is True then the relationship will be directional.
    #
    def relation_create( self, parent, child, directed = False ):

        d = 'N'
        if directed:
            d = 'Y'

        with self.conn as c:
            c.execute( '''INSERT INTO relationship (parent, child, directed)
                VALUES (?,?,?)''', (parent, child, d) )
        return


    # Remove the relationship between two entities. The entities themselves
    # are unchanged.
    #
    def relation_remove( self, parent, child ):

        with self.conn as c:
            c.execute( '''DELETE FROM relationship WHERE parent=? AND child=?''',
                       (parent, child) )
        return


    # Change the valued of the directivity for a relationship.
    #
    def relation_change_directivity( self, parent, child, directed ):

        d = 'N'
        if directed:
            d = 'Y'

        with self.conn as c:
            c.execute( '''UPDATE relationship SET directed=? WHERE
                parent=? AND child=?''', (parent, child, d) )
        return
    

    ##
    ## Search and retrieval
    ##

    # Search the KB for entities matching the search phrase.
    # Return a result set containing the matches.
    #
    def search( self, phrase, search_type = None ):

        search_phrase = '%' + phrase + '%'
        
        with self.conn as c:

            cur = c.execute( '''
    SELECT id, name, content, STRFTIME( '%Y-%m-%d', modified ) AS modified FROM (
        SELECT id, name, "Category" AS content, modified
            FROM kb_entity WHERE name LIKE ? AND type = 'C'
        UNION
        SELECT id, COALESCE(name, "Empty") AS name, SUBSTR(content,1,45) AS content,
               modified
            FROM frag WHERE content LIKE ? OR name LIKE ?
    ) ORDER BY modified DESC
            ''', (search_phrase, search_phrase, search_phrase) )
            
        return cur


    # Return only the name of an entity.
    #
    def entity_name( self, entity_id ):

        rec = None
        with self.conn as c:

            cur = c.execute( '''
            SELECT name FROM kb_entity WHERE id = ?
            ''', (entity_id,) )

            rec = cur.fetchone()[0]

        return rec
    

    # Search for an entity matching a UUID and return it.
    # The return value is a tuple where the first element is the kind of
    # entity: 'K' for a knowledge fragment and 'C' for a catagory. The
    # second element is the entity name, and the third is the respective data.
    #
    def retrieve_entity( self, entity_id ):

        rec = None
        
        with self.conn as c:

            e_type = self.retrieve_entity_type( entity_id )

            if e_type == 'K':

                row = self.fetch_kfrag( entity_id )
                rec = ( 'K', row['name'], row )
            else: # a category

                rows = self.retrieve_cat_contents( entity_id )
                e_name = self.entity_name( entity_id )
                rec = ( 'C', e_name, rows )
        
        return rec


    # Return the row for a knowledge fragment.
    #
    def fetch_kfrag( self, entity_id ):

        rec = None
        with self.conn as c:

            cur = c.execute( '''
            SELECT 'K' AS type, name, created, modified, origin, content,
                 binary, content_type, binary_type
              FROM frag WHERE id = ?
            ''', (entity_id,) )

            rec = cur.fetchone()

        return rec


    # Return a list of child relationships for a parent entity.
    #
    def retrieve_relations( self, entity_id ):

        rec = None
        with self.conn as c:

            cur = c.execute( '''
              SELECT child, directed, name
                FROM relationship r JOIN kb_entity e
                  ON r.child = e.id
                WHERE parent = ?
            ''', (entity_id,) )

            rec = cur.fetchall()
            
        return rec


    # Return the type of an entity.
    #
    def retrieve_entity_type( self, entity_id ):

        rec = None
        
        with self.conn as c:

            cur = c.execute( '''
                SELECT name, type FROM kb_entity
                   WHERE id = ?
            ''', (entity_id,) )

            rec = cur.fetchone()[1]
        return rec


    # Return a list of the contents of a category.
    #
    def retrieve_cat_contents( self, entity_id ):

        rec = None

        with self.conn as c:

            cur = c.execute( '''
              SELECT contains, name, ordering
                FROM category c JOIN kb_entity e
                  ON c.contains = e.id
                WHERE c.id = ?
            ''', (entity_id,) )

            rec = cur.fetchall()
            
        return rec
    

# Return a random UUID value as a hex string.
#
def uuid_create():

    return uuid.uuid4().hex


