# Copyright (C) 2018, 2019 Nathan Matthew Miller

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import tkinter as tk

import io
from enum import Enum

class Type(Enum):
    TEXT = 0
    IMG  = 1


# Take Tk's index value, which is formatted as a string, and return a
# tuple with the line and character positions.
#
def index_to_int( index ):

    (line, pos) = index.split( '.' )
    return ( int(line), int(pos) )


class Model:

    def __init__( self, view ):

        self.view = view
        self.doc  = []
        self.doc.append( [ Type.TEXT, "", None ] )

        self.ctrl_key_dispatch = {
            'BackSpace':self.backspace,
            'Delete':   self.delete,
            'Tab':      self.tab,
            'Return':   self.new_line
            }
        
        return


    # Take the content text (defaulting to org-mode format) and create the
    # model for that text.
    #
    # The model created is simple: a list containing all of the lines.
    # Each entry in the list is another list which is the node of a singly
    # linked list. The last element of the node list is a reference to
    # the next node or None if no other node.
    #
    def create_model( self, content, format="org" ):

        self.format = format
        self.doc.clear()

        fd = io.StringIO( content )

        line = fd.readline()
        while len(line) > 0:

            model_line = [ Type.TEXT, line, None ]
            self.doc.append( model_line )

            line = fd.readline()

        fd.close()
        print( f"Loaded model with {len(self.doc)} lines" )
        return


    # Use index information from the Text widget to return the line, image,
    # or other segment from the model.
    # Return a tuple containing the index into the line and the segment
    # containing the line.
    # TODO: Impliment the calculation to account for images in the line
    #
    def get_model_segment( self, line_no, point ):

        model_line = self.doc[ line_no-1 ]
        
        # Find line segment containing the point
        # Calculate the index into the line segment
        i = point
        segment = model_line

        return (i, segment)


    # For a line, return the last line segement in the list.
    #
    def get_last_segment( self, line_no ):

        model_line = self.doc[ line_no-1 ]

        link = model_line[ len(model_line)-1 ]
        while link != None:

            model_line = link
            link = model_line[ len(model_line)-1 ]
            
        return model_line


    def remove_line( self, line_no ):

        del self.doc[line_no-1]
        return

    
    def backspace( self, line_no, point ):

        # If there's currently a selection, only delete the selected text.
        if self.valid_range:
            self.delete_selection( self.view.text )
            # Do not process this event further, otherwise a character is
            # removed from the widget which shouldn't be.
            return 'break'

        # If the cursor in the text widget is on the last line and
        # backspace is pressed, there may not be an actual line to remove
        # from the model.
        if line_no > len(self.doc):
            return
        
        (i, segment) = self.get_model_segment( line_no, point )
        line_seg = segment[1]
        
        if i == 0:
            if line_no != 1:

                # Append this line to the line above it.
                prior_seg = self.get_last_segment( line_no-1 )
                edited_line = prior_seg[1][:-1] + line_seg
                prior_seg[1] = edited_line

                # Remove the line line_no from the document.
                self.remove_line( line_no )                
        else:
        
            edited_line = line_seg[:i-1] + line_seg[i:]
            segment[1] = edited_line
            
        return


    def delete( self, line_no, point ):

        # If there's currently a selection, only delete the selected text.
        if self.valid_range:
            self.delete_selection( self.view.text )
            # Do not process this event further, otherwise a character is
            # removed from the widget which shouldn't be.
            return 'break'

        # If the cursor in the text widget is on the last line and
        # delete is pressed, there may not be an actual line to edit
        # in the model.
        if line_no > len(self.doc):
            return

        (i, segment) = self.get_model_segment( line_no, point )
        line_seg = segment[1]

        if i+1 == len( line_seg ) and segment[len(segment)-1] == None:
            # then we're at the end of the line
            if line_no != len( self.doc ):
                next_line = self.doc[ line_no ][1]
                edited_line = line_seg[:-1] + next_line
                segment[1] = edited_line
                self.doc.pop( line_no )
                
        else:
            edited_line = line_seg[:i] + line_seg[i+1:]
            segment[1] = edited_line
                        
        return


    def new_line( self, line_no, point ):

        if line_no > len( self.doc ):
            self.doc.append( [ Type.TEXT, "\n", None ] )
            
        else:
            (i, segment) = self.get_model_segment( line_no, point )
            line_seg = segment[1]

            # Split the line and push the right-most remainder down one line.
            current_line = line_seg[:i] + "\n"
            segment[1] = current_line
            next_line = line_seg[i:]

            if len(next_line) > 0:
                self.doc.insert( line_no, [ Type.TEXT, next_line, None ] )

        return

    
    def tab( self, line_no, point ):

        return 'break'


    def insert_char( self, line_no, point, char ):

        # Is the Text widget cursor position on the line following the last
        # actual line of content?
        if line_no == len( self.doc )+1:
            self.doc.append( [ Type.TEXT, char, None ] )
            return
        
        (i, segment) = self.get_model_segment( line_no, point )

        line_seg = segment[1]
        edited_line = line_seg[:i] + char + line_seg[i:]
        segment[1] = edited_line

        return


    # Load the model into the Text widget.
    #
    def load_model( self, entity_name ):

        for line in self.doc:
            self.view.text.insert( tk.END, line[1] )

        if entity_name != None:
            self.view.set_title( entity_name )
        return
    

    # Load a KB entity into the Text widget.
    # Assumption is that the Text widget is empty.
    #
    def load_entity( self, e_name, entity_details, relations ):

        self.create_model( entity_details[2]['content'] )
        self.load_model( e_name )

        for r in relations:
            self.view.rel_view.add_relation( r['child'], r['name'] )
        return


    # Return the textual content of the model.
    # TODO: handle lines with embedded images.
    #
    def get_content( self ):

        content = ""

        for line in self.doc:
            content = content + line[1]
            
        return content


    # Return the title of this entity.
    #
    def get_title( self ):
        return self.view.entity_name.get()
    

    # A range of text in the widget is highlighed.
    #
    def set_range( self, range_begin, range_end ):

        self.range_begin = range_begin
        self.range_end   = range_end
        self.valid_range = True
        return


    # There is currently no selected range in the widget.
    #
    def clear_range( self ):

        self.valid_range = False
        return


    # A keypress occurred. Update the model, and if necessary the view.
    #
    def key_press( self, index, event ):

        (line, point) = index_to_int( index )
        #print( "%d.%d %s" % (line, point, event.keysym))
        rtn = None

        if event.keysym_num > 0xFF00:

            # Handle control or non-printing keys specially
            try:
                rtn = self.ctrl_key_dispatch[ event.keysym ]( line, point )
            except KeyError:
                None
            
        else: # Some printable character

            rtn = self.insert_char( line, point, event.char )
            
        return rtn


    # If the line of the model doesn't have a terminating newline charactor
    # then add one. Return True if it was necessary to add a newline.
    #
    def check_add_newline( self, line_no ):

        line_seg = self.get_last_segment( line_no )

        if line_seg[0] == Type.TEXT:

            if line_seg[1][-1] != "\n":
                line_seg[1] = line_seg[1] + "\n"
                return True
        else:

            line_seg[ len(line_seg)-1 ] = [ Type.TEXT, "\n", None ]
            return True

        return False
    

    # A line in the model has been updated, update the view as well.
    #
    def update_view( self, line_no ):

        # Remove the text of the line (except for the newline) and insert
        # the current text from the model (except for the newline).
        self.view.text.delete( f"{line_no}.0", f"{line_no}.lineend" )
        line_seg = self.doc[ line_no-1 ]
        self.view.text.insert( f"{line_no}.0", line_seg[1].rstrip() )
        return
    

    # Update the view by adding a link to another entity.
    # The model is just an intermediary here...
    #
    def add_relationship_link( self, link_id, entity ):

        self.view.rel_view.add_relation( link_id, entity[1] )
        return


    # Remove a selected area of text from the model.
    #
    def delete_selection( self, widget ):

        if not self.valid_range:
            return # no selection present

        widget.delete( self.range_begin, self.range_end )

        (bline, bpos) = index_to_int( "%s" % self.range_begin )
        (eline, epos) = index_to_int( "%s" % self.range_end   )

        #print( f"{bline}.{bpos}")
        #print( f"{eline}.{epos}")

        (m_pos, a_seg) = self.get_model_segment( bline, bpos )
        (n_pos, b_seg) = self.get_model_segment( eline, epos )

        left_str = a_seg[1][ :m_pos ]
        i = bline
        del_list = []

        # Get the right most line segment and also any lines spanned by the
        # selection which will need to be deleted.
        while True:
            if i == eline:
                right_str = b_seg[1][ n_pos: ]
                break
            else:
                i = i + 1
                del_list.append( i )

        # Update the line segment.
        edited_line = left_str + right_str
        a_seg[1] = edited_line

        del_list.reverse()
        for line_index in del_list:
            self.remove_line( line_index )

        #print( edited_line )
        return


    # Take the contents of the clipboard and insert it into the model.
    #
    def insert_paste( self, widget ):

        clip_text = widget.clipboard_get()

        point = widget.index( "insert" )
        print( "%s" % point )
        (line, pos) = index_to_int( "%s" % point )

        # Split the line containing the insertion point and insert the
        # pasted text.

        (i, seg) = self.get_model_segment( line, pos )
        left_str  = seg[1][ :i ]
        right_str = seg[1][ i: ]

        #print( f".{left_str}.")
        #print( f".{right_str}.")

        fd = io.StringIO( clip_text )
        left_str = left_str + fd.readline()

        n = line
        model_line = None
        while True:

            text_line = fd.readline()
            if len(text_line) > 0:
                model_line = [ Type.TEXT, text_line, None ]
                self.doc.insert( n, model_line )
                n = n + 1
            else:
                break

        fd.close()
        if model_line == None:

            seg[1] = left_str + right_str
        else:
            seg[1] = left_str
            model_line[1] = model_line[1] + right_str
        return
