# Copyright (C) 2018, 2019 Nathan Matthew Miller

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import tkinter as tk
from tkinter import messagebox
from TkGUI import RelationshipView, LinkView

class View:

    def __init__( self, top, controller, entity_type ):

        self.ctrl = controller
        self.top  = top
        self.entity_type = entity_type

        # Entity name text entry
        self.entity_name = tk.Entry( top )
        self.entity_name.pack( anchor=tk.NW, fill=tk.X )
        self.entity_name.bind( "<KeyRelease>", self.name_updated )
        
        # An area to show status messages.
        self.status_bar = tk.Label( top, relief=tk.SUNKEN )
        self.status_bar.pack( side=tk.BOTTOM, fill=tk.X )

        if entity_type == 'K':
            
            # Text widget to display the KB entity, including scrollbars.
            tframe = tk.Frame( top )
            yscroll = tk.Scrollbar( tframe )
            yscroll.pack( side=tk.RIGHT, fill=tk.Y )

            xscroll = tk.Scrollbar( tframe, orient=tk.HORIZONTAL )
            xscroll.pack( side=tk.BOTTOM, fill=tk.X )

            self.text = tk.Text( tframe, wrap=tk.WORD,
                                 xscrollcommand=xscroll.set,
                                 yscrollcommand=yscroll.set )

            xscroll.config( command=self.text.xview )
            yscroll.config( command=self.text.yview )

            # Without these arguments to pack() the Text widget will not stretch
            # or shrink to fill the frame.
            self.text.pack( expand=True, fill='both' )
            tframe.pack()
        else:

            self.text = LinkView.View( self.top, self.ctrl, text_height=20 )
            self.text.pack()

            self.text.set_remove_method( self.ctrl.remove_from_category )
        
        # Display entity relationships in a Text widget below the main
        # Text widget.
        self.rel_view = RelationshipView.ViewBox( self.top, self.ctrl )
        self.rel_view.pack()

        if entity_type == 'K':
            # Bind event handlers which will gather information to keep the view
            # and model in sync.
            self.text.bindtags(  ('pre-class-binding', 'Text') )
            self.text.bind_class( "pre-class-binding", "<KeyPress>",
                                  self.ctrl.pre_handler )
            self.text.bind_class( "pre-class-binding", "<ButtonRelease-1>",
                                  self.ctrl.pre_handler )

            self.text.bind_class( "pre-class-binding", "<<Paste>>",
                                  self.ctrl.paste_handler )
            self.text.bind_class( "pre-class-binding", "<<Cut>>",
                                  self.ctrl.cut_handler )
            
        self.menuBar = tk.Menu( top )
        top['menu'] = self.menuBar

        self.subMenu = tk.Menu( self.menuBar )
        self.menuBar.add_cascade( label='KB Entity',  menu=self.subMenu )
        
        self.subMenu.add_command( label='Save',
                                  command=self.save_handler )
        self.subMenu.add_command( label='About',
                                  command=self.about_handler )

        self.subMenu.entryconfigure( "Save", state="disabled" )

        # Setup the structure menu and its items
        s_menu = tk.Menu( self.menuBar )
        self.menuBar.add_cascade( label='Structure', menu=s_menu )

        s_menu.add_command( label='Store ref', accelerator='Ctrl-r',
                            command=lambda: self.ctrl.store_entity_id(None) )
        s_menu.add_command( label='Link to', accelerator='Ctrl-l',
                            command=lambda: self.ctrl.link_to_entity(None) )

        if entity_type == 'C':
            s_menu.add_command( label='Category add', accelerator='Ctrl-a',
                           command=lambda: self.ctrl.add_to_category(None) )

        self.set_command_keys()
        
        self.text.after( 1000, self.is_modified )
        return
    

    def about_handler( self ):

        messagebox.showinfo( "About", "Exocortex knowledge store" )
        return


    # Called when the Save menu item is selected.
    #
    def save_handler( self ):

        self.ctrl.save_entity()
        self.subMenu.entryconfigure( "Save", state="disabled" )
        return


    # Called to check if the Text widget has been modified.
    #
    def is_modified( self ):

        if self.text.edit_modified():
            
            self.subMenu.entryconfigure( "Save", state="normal" )
            self.text.edit_modified( False )

        self.text.after( 1000, self.is_modified )
        return


    def clear_modified( self ):

        self.text.edit_modified( False )
        return


    # Set the keys and key combos for the editor.
    # Due to the Text bind class, the bindings must be set for both it and
    # the TopLevel widget.
    #
    def set_command_keys( self ):

        self.top.bind( "<Control-r>", self.ctrl.store_entity_id )
        self.top.bind( "<Control-l>", self.ctrl.link_to_entity )

        self.rel_view.bind( "<Control-r>", self.ctrl.store_entity_id )
        self.rel_view.bind( "<Control-l>", self.ctrl.link_to_entity )
        
        if self.entity_type == 'K':
            self.text.bind_class( "pre-class-binding", "<Control-r>",
                                  self.ctrl.store_entity_id )
            self.text.bind_class( "pre-class-binding", "<Control-l>",
                                  self.ctrl.link_to_entity )
        else:
            self.text.bind( "<Control-r>", self.ctrl.store_entity_id )
            self.text.bind( "<Control-l>", self.ctrl.link_to_entity )

            self.rel_view.bind( "<Control-a>", self.ctrl.add_to_category )
            self.top.bind(      "<Control-a>", self.ctrl.add_to_category )
            self.text.bind(     "<Control-a>", self.ctrl.add_to_category )
        return


    def set_title( self, title ):

        self.entity_name.insert( 0, title )

        if self.entity_type == 'C':
            title = title + " - Category"
        
        self.top.title( title )
        return


    def name_updated( self, event ):

        title = self.entity_name.get()

        if self.entity_type == 'C':
            title = title + " - Category"
            
        self.top.title( title )

        self.text.edit_modified( True )
        return
    
