# Copyright (C) 2018, 2019 Nathan Matthew Miller

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import tkinter as tk
from tkinter.font import *

# Display links, whether for relationships or categories, in a Text widget.
#
# After creation, methods are set up so the correct action is taken by the
# controller when an action occurs.

# Take Tk's index value, which is formatted as a string, and return a
# tuple with the line and character positions.
#
def index_to_int( index ):

    (line, pos) = index.split( '.' )
    return ( int(line), int(pos) )


class View( tk.Frame ):

    def __init__( self, top, ctrl, text_height=4 ):

        super().__init__( top )

        self.ctrl = ctrl
        
        ryscroll = tk.Scrollbar( self )
        ryscroll.pack( side=tk.RIGHT, fill=tk.Y )
        self.text = tk.Text( self, yscrollcommand=ryscroll.set,
                                   height=text_height )
        ryscroll.config( command=self.text.yview )
        self.text.pack( expand=True, fill=tk.Y )

        # Prevent inadvertent editting of the widget.
        # This doesn't work. Binding on KeyPress prevents Ctrl-a and others
        # from working... Must be a better way.
        self.text.bind( "<KeyRelease>", lambda event: "break" )

        self.xfont = Font( family='Helvetica', size=-12, weight='bold' )

        self.links = [] # List of all displayed links
        return


    def set_remove_method( self, method ):
        self.remove_link = method
        return
    

    def mouse_enter( self, event ):
        self.text.config( cursor="hand2" )
        return
    

    def mouse_leave( self, event ):
        self.text.config( cursor="xterm" )
        return


    def link_delete( self, event ):
        self.text.config( cursor="xterm" ) # reset the cursor
        
        i = event.widget.index( f"@{event.x},{event.y}" )
        (line,pos) = index_to_int( i )

        id = self.links[ line-1 ][0]
        self.remove_link( id )
        
        del self.links[ line-1 ]
        self.text.delete( f"{line}.0", f"{line+1}.0" )
        
        return


    def link_follow( self, event ):
        i = event.widget.index( f"@{event.x},{event.y}" )
        (line,pos) = index_to_int( i )

        id = self.links[ line-1 ]
        self.ctrl.show_entity( id[0] )
        return


    def show_relation( self, line, id, title ):

        self.text.insert( f"{line}.0", "X " + title + "\n" )
        self.text.tag_add( "HOVER", f"{line}.0", f"{line}.1" )
        
        self.text.tag_config( "HOVER", font=self.xfont, background="red" )
        self.text.tag_bind( "HOVER", "<Enter>", self.mouse_enter )
        self.text.tag_bind( "HOVER", "<Leave>", self.mouse_leave )
        self.text.tag_bind( "HOVER", "<Button-1>", self.link_delete )

        self.text.tag_add( "LINK", f"{line}.2", f"{line}.end" )
        self.text.tag_config( "LINK", foreground="blue", underline=1 )
        self.text.tag_bind( "LINK", "<Enter>", self.mouse_enter )
        self.text.tag_bind( "LINK", "<Leave>", self.mouse_leave )
        self.text.tag_bind( "LINK", "<Button-1>", self.link_follow )
        return


    # Add a new relationship to the view, which is indexed by the id value.
    #
    def add_relation( self, id, title ):

        self.links.append( (id, title) )
        line = len( self.links )
        
        self.show_relation( line, id, title )
        return
    

    def edit_modified( self, modified=None ):

        return self.text.edit_modified( modified )
