# Copyright (C) 2018, 2019 Nathan Matthew Miller

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import re
import tkinter as tk
from tkinter import messagebox
from tkinter.font import *

class View:

    def __init__( self, root, controller ):

        self.ctrl = controller
        
        frame = tk.Frame( root )
        frame.pack( side=tk.TOP )

        # Create the search bar across the top.
        #
        tk.Label( frame, text="Keywords: " ).pack( side=tk.LEFT, anchor=tk.W )

        entry = tk.Entry( frame, width=30 )
        entry.pack( side=tk.LEFT )
        entry.focus_set()
        entry.bind( "<KeyPress-Return>", self.do_search )
        self.entry = entry
        
        search = tk.Button( frame, text="Search", command=self.do_search )
        search.pack( side=tk.LEFT )
        search.bind( "<KeyPress-Return>", self.do_search )

        # Another frame holds the ListBox which shows search results.
        #
        frame2 = tk.Frame( root )
        frame2.pack( side=tk.BOTTOM )

        results = tk.Listbox( frame2, width=90, height=50,
                  selectmode=tk.SINGLE, exportselection=0 )

        results.pack( anchor=tk.S )
        results.bind( "<Double-Button-1>", self.result_selected )
        results.bind( "<KeyPress-Return>", self.result_selected )
        self.results_index = 0
        results.bind( "<Up>", self.move_selection_up )
        results.bind( "<Down>", self.move_selection_down )

        self.results = results
        self.matches = None

        # Set keyboard shortcuts
        for w in [frame, entry, frame2, results]:
            w.bind( "<Control-f>", lambda e: self.ctrl.new_frag() )
            w.bind( "<Control-c>", lambda e: self.ctrl.new_cat() )

        # Menu setup
        #
        self.menuBar = tk.Menu( root )
        root[ 'menu' ] = self.menuBar

        # File menu
        self.fileMenu = tk.Menu( self.menuBar )
        self.menuBar.add_cascade( label='File', menu=self.fileMenu )

        self.fileMenu.add_command( label='Import' )
        self.fileMenu.entryconfigure( 'Import', state='disabled' )
        
        self.fileMenu.add_command( label='Exit', command=root.quit )

        # New entity menu
        self.entMenu = tk.Menu( self.menuBar )
        self.menuBar.add_cascade( label='New', menu=self.entMenu )

        self.entMenu.add_command( label='Fragment', command=self.ctrl.new_frag,
                                  accelerator='Ctrl-f' )
        self.entMenu.add_command( label='Category', command=self.ctrl.new_cat,
                                  accelerator='Ctrl-c' )

        return
    

    def move_selection_up( self, event ):

        i    = self.results_index

        if i > 0:
            self.results.select_clear( i )
            self.results_index = i - 1
            self.results.select_set( self.results_index )
        return


    def move_selection_down( self, event ):

        size = self.results.size()
        i    = self.results_index

        if i < size:
            self.results.select_clear( i )
            self.results_index = i + 1
            self.results.select_set( self.results_index )
        return


    # An item in the results listbox has been either double clicked or
    # selected using the Return key.
    #
    def result_selected( self, event ):

        i = self.results.curselection()[0]
        self.ctrl.show_entity( self.matches[i][0] )
        return
    

    def do_search( self, event=None ):

        phrase = self.entry.get()

        self.results_index = 0
        
        if len(phrase) == 0:

            messagebox.showerror( "Error", "No search phrase" )
            return

        matches = self.ctrl.search_kb( phrase )

        if len(matches) == 0:

            self.results.delete( 0, tk.END )
            messagebox.showinfo( "Search results", "No matches found" )
            return

        self.results.delete( 0, tk.END )
        self.results_index = 0
        strip_org = re.compile( "^[\*\s]*(.+)$" )
        
        for match in matches:
            body = match[2].replace("\n", " ")
            m = strip_org.match( body )
            if m != None:
                body = m[1]
                
            s = '{:<35}|{:<40}| {}'.format( match[1][:35],
                                            body[:40],
                                            match[3] )
            self.results.insert( tk.END, s )
            
        self.matches = matches
