#!/usr/bin/env python3

# Initialize the knowledge base configuration file and database file.

# Copyright (C) 2018, 2019 Nathan Matthew Miller

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sys
import os.path
import sqlite3
import platform
from os.path import join

import tkinter as tk
from tkinter.font import *
from tkinter import messagebox, filedialog

import KB.Core as core

setup_dir = ""

def exit_gui():
    global root
    root.quit()


class Setup:


    def __init__( self, frame ):
        self.f = frame

        if platform.system() == "Windows":
            self.platform = 'windows'
        else:
            self.platform = 'unix'

        self.REG_PATH = r'Software\Exocortex'
        return


    def select_dir_dialog( self ):

        self.setup_dir = filedialog.askdirectory( title="Select Folder" )
        if self.setup_dir != None:
            self.schema_button.configure( state="normal" )
        return


    def pick_schema( self ):

        self.sql_file = filedialog.askopenfilename( filetypes=[("sql files", "*.sql")] )
        
        if self.sql_file != None:
            self.init_button.configure( state="normal" )
        return


    # Create the configuration file (or Windows registry entry) and create
    # the database file.
    #
    def setup_db( self ):

        self.create_configuration()
        self.initialize_db()

        messagebox.showinfo( "Exocortex Setup", "Database setup successful!",
                             parent=self.f )
        return


    def create_configuration( self ):

        self.db_file_path = join( self.setup_dir, "exocortex.db" )

        if self.platform == 'windows':

            self.windows_set_reg_key()
        else:
            
            # Create the configuration file
            home_path = os.path.expanduser( '~/' )
            config_file = home_path + '/.exocortex'

            with open( config_file, 'w' ) as fd:
                fd.write( "[databases]\n" +
                          f"main = {self.db_file_path}\n" )
        return


    # Set the registry key which tells the application where to find the
    # database file.
    #
    def windows_set_reg_key( self ):

        import winreg

        try:
            winreg.CreateKey( winreg.HKEY_CURRENT_USER , self.REG_PATH )
            registry_key = winreg.OpenKey( winreg.HKEY_CURRENT_USER,
                                    self.REG_PATH, 0, winreg.KEY_WRITE )
            
            winreg.SetValueEx( registry_key, r'main', 0, winreg.REG_SZ,
                               self.db_file_path )
            winreg.CloseKey( registry_key )
            
        except OSError as error:
            messagebox.showerror( "Error!", "windows_set_reg_key(): " + str(error),
                                  parent=self.f )
        return


    # Initialize the database file
    #
    def initialize_db( self ):

        conn = sqlite3.connect( self.db_file_path )
        with open( self.sql_file, 'r' ) as fd:
            sql = fd.read()
        conn.executescript( sql )
        conn.close()

        return
    

    def load_widgets( self ):

        l1 = tk.Label( self.f,
                       text="1. Choose where to put the database file: " )
        config_button = tk.Button( self.f, text="Pick Folder",
                                   command=self.select_dir_dialog )
        l1.grid( column=0, row=0, sticky=tk.W )
        config_button.grid( column=1, row=0, sticky=tk.W+tk.E )

        
        l2 = tk.Label( self.f, text="2. Select the DB schema to load: " )
        schema_button = tk.Button( self.f, text="Pick Schema",
                                   command=self.pick_schema )
        l2.grid( column=0, row=1, sticky=tk.W )
        schema_button.grid( column=1, row=1, sticky=tk.W+tk.E )
        
        l3 = tk.Label( self.f, text="3. Initialize the database: " )
        init_button = tk.Button( self.f, text="Create",
                                 command=self.setup_db )
        l3.grid( column=0, row=2, sticky=tk.W )
        init_button.grid( column=1, row=2, sticky=tk.W+tk.E )

        schema_button.configure( state="disable" )
        init_button.configure( state="disable" )

        self.schema_button = schema_button
        self.init_button = init_button
        return


# Main program starts

root = tk.Tk()
root.wm_protocol( 'WM_DELETE_WINDOW', exit_gui )
root.wm_title( "Exocortex DB Initialization" )

frame = tk.Frame( root, padx=5, pady=5 )
frame.pack()

setup = Setup( frame )
setup.load_widgets()

tk.mainloop()
sys.exit( 0 )
